//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see http://www.gnu.org/licenses/.
// 

#include "UDPBasicAppCam.h"


#include "UDPBasicApp.h"

#include "InterfaceTableAccess.h"
#include "IPvXAddressResolver.h"
#include "NodeOperations.h"
#include "UDPControlInfo_m.h"
#include "DecisionMaker.h"
#include "random.h"

#ifdef WITH_IPv4
#include "IRoutingTable.h"
#include "RoutingTableAccess.h"
#endif

#ifdef WITH_IPv6
#include "RoutingTable6.h"
#include "RoutingTable6Access.h"
#endif

Define_Module(UDPBasicAppCam);

simsignal_t UDPBasicAppCam::rcvdPkSignal = registerSignal("rcvdPkSignal");
simsignal_t UDPBasicAppCam::sentCamSignal = registerSignal("sentCamSignal");



UDPBasicAppCam::UDPBasicAppCam()
{

    selfMsg = NULL;
    host = NULL;
    nodeStatus = NULL;
    mobility = NULL;
    routingTable = NULL;
//    outInterfaceId=101;
    outputInterface = -1;
    outputInterfaceMulticastBroadcast.clear();


    std::cout <<"1"<<endl;
}

UDPBasicAppCam::~UDPBasicAppCam() {
    UDPBasicApp::cancelAndDelete(selfMsg);
    std::cout <<"2"<<endl;
}


void UDPBasicAppCam::initialize(int stage)
{
    std::cout<<" initialize "<<std::endl;
    UDPBasicApp::initialize(stage);

    if (stage == 0) {

        numCamSent = 0;
        numCamReceived = 0;
        WATCH(numCamSent);
        WATCH(numCamReceived);

        localPort = par("localPort");
        destPort = par("destPort");
        camInterval = par("camInterval");
        T_GenCamMin = par("T_GenCamMin");
        T_GenCamMax = par("T_GenCamMax");
        camTimer = new cMessage("camTimer");

        destAddrRNG = par("destAddrRNG");




        host = getContainingNode(this);
        nodeStatus = dynamic_cast<NodeStatus *>(host->getSubmodule("status"));
        mobility = check_and_cast<IMobility *>(host->getSubmodule("mobility"));
        routingTable = check_and_cast<IRoutingTable *>(getModuleByPath(par("routingTableModule")));

        carCamGeneration(host, mobility);



    } else if (stage == 3) {
            if (par("configureInInit").boolValue())
                initialConfiguration();
         }
    std::cout <<"3"<<endl;
}


IPvXAddress UDPBasicAppCam::getSelfAddress() const
{
    return routingTable->getRouterId();
}


void UDPBasicAppCam::finish() {
   UDPBasicApp::finish();
    std::cout <<"4"<<endl;
}

void UDPBasicAppCam::setSocketOptions() {
   UDPBasicApp::setSocketOptions();

    std::cout <<"5"<<endl;
}

IPvXAddress UDPBasicAppCam::chooseDestAddr() {
//    UDPBasicApp::chooseDestAddr();

    if (destAddresses.size() == 1)
          return destAddresses[0];

      int k = genk_intrand(destAddrRNG, destAddresses.size());
      return destAddresses[k];

    std::cout <<"6"<<endl;
}

void UDPBasicAppCam::sendPacket() {
    UDPBasicApp::sendPacket();


    std::cout <<"7"<<endl;
}


void UDPBasicAppCam::processStart() {
  UDPBasicApp::processStart();



//  socket.setOutputGate(gate("udpOut"));
//  socket.bind(localPort);

  std::cout <<" UDPBasicAppCam::processStart()"<<endl;
  std::cout <<" UDPBasicAppCam::processStart()"<<endl;
  std::cout <<" UDPBasicAppCam::processStart()"<<endl;

}

void UDPBasicAppCam::processSend() {
   UDPBasicApp::processSend();



    std::cout <<"9"<<endl;
}


void UDPBasicAppCam::processStop() {
//   UDPBasicApp::processStop();
   socket.close();


    std::cout <<"10"<<endl;
}


void UDPBasicAppCam::processPacket(cPacket *pk) {
 UDPBasicApp::processPacket(pk);

 emit(rcvdPkSignal, pk);
     EV << "Received CAM packet: " << UDPSocket::getReceivedPacketInfo(pk) << endl;
     delete pk;
     numCamReceived++;

    std::cout <<"12"<<endl;

}

bool UDPBasicAppCam::handleNodeStart(IDoneCallback *doneCallback) {
   UDPBasicApp::handleNodeStart(doneCallback);



    std::cout <<"13"<<endl;
}

bool UDPBasicAppCam::handleNodeShutdown(IDoneCallback *doneCallback) {
   UDPBasicApp::handleNodeShutdown(doneCallback);


}

void UDPBasicAppCam::handleNodeCrash() {
    UDPBasicApp::handleNodeCrash();

    std::cout <<"15"<<endl;
}

basicCamContainer * UDPBasicAppCam::carCamGeneration(cModule *host, IMobility *mobility){
    this->host = host;
    this->mobility = mobility;

    basicCamContainer * mcam = new basicCamContainer();
    mcam->setSpeed(mobility->getCurrentSpeed());
    mcam->setPosition(mobility->getCurrentPosition());
    mcam->setCreateTime(simTime());
    mcam->setAddress(getSelfAddress());
    mcam->setHeading(getVectorAngle(mobility->getCurrentPosition()));


    basicCamContainer * lastmcam=mcam;

    IPvXAddress address=mcam->getAddress();
    positionTableforMobilityCam.setSpeed(address,mcam->getPosition(),mcam->getSpeed());
    positionTableforMobilityCam.getSpeed(address);

    positionTableforMobilityCam.setPosition(address,mcam->getPosition(),mcam->getSpeed());
    positionTableforMobilityCam.getCreateTime(address);

    positionTableforMobilityCam.setHeading(address,mcam->getHeading());

    std::cout <<"hostname"<<host->getFullName()<<endl;
    std::cout <<"hostname"<<host->getFullName()<<endl;
    std::cout <<"hostname"<<host->getFullName()<<endl;
    std::cout <<"hostname"<<host->getFullName()<<endl;

    return lastmcam;
}


// check conditions  speed, position ,heading
bool UDPBasicAppCam::checkSpeed(Coord curSpeed,Coord preSpeed){

  static double speedThreshold=0.5;
  double speedDel = (curSpeed-preSpeed).length();

  return speedDel>speedThreshold;
}

bool UDPBasicAppCam::checkPosition( Coord curPosition,Coord prePosition){

    static double positionThreshold=1;

    double positionDel = (curPosition-prePosition).length();

    return positionDel>positionThreshold;
}

bool UDPBasicAppCam::checkHeading( Coord curPosition,Coord prePosition){

    static double headingThreshold=1;

    double preHeading=getVectorAngle(prePosition);
    double curHeading=getVectorAngle(curPosition);

    double headingDel = abs(preHeading-curHeading)*180;

    return headingDel>headingThreshold;
}

double UDPBasicAppCam::getVectorAngle(Coord vector)
{
    double angle = atan2(vector.y, vector.x);
    adjustVectorAngle(angle);
    return angle;
}
double UDPBasicAppCam::adjustVectorAngle(double angle)
{

    if (angle < 0)
        angle += 2 * PI;
    if (angle > PI)
        angle=angle- PI;
    if (angle > PI/2)
        angle=PI-angle;

    return angle;
}


void UDPBasicAppCam::CheckCamCondition( IPvXAddress selfAddress){
    std::cout<<"  CheckCamCondition               begin begin begin               " <<endl;

    selfAddress = getSelfAddress();

    // lastmcam in PositionTable
    Coord prePosition=positionTableforMobilityCam.getPosition(selfAddress);
    Coord preSpeed=positionTableforMobilityCam.getSpeed(selfAddress);
   simtime_t T_preGenCam=positionTableforMobilityCam.getCreateTime(selfAddress);

    // current generation Cam
    basicCamContainer * curcam = new basicCamContainer();
    curcam->setSpeed(mobility->getCurrentSpeed());
    curcam->setPosition(mobility->getCurrentPosition());
    curcam->setCreateTime(simTime());
    curcam->setAddress(selfAddress);

    curcam->setHeading(getVectorAngle(mobility->getCurrentPosition()));


   Coord curPosition= curcam->getPosition();
   Coord curSpeed=curcam->getSpeed();


   std::cout<<"  selfAddress: "<< selfAddress <<endl;
   std::cout<<"  prePosition: "<< prePosition <<endl;
   std::cout<<"  curPosition: "<< curPosition <<endl;
   std::cout<<"  preSpeed: "<< preSpeed <<endl;
   std::cout<<"  curSpeed: "<< curSpeed <<endl;

   // Time Condition
   simtime_t T_GenCamDel=simTime()-T_preGenCam;
   simtime_t T_GenCamDcc;

   //it should be by DCC  Algorithm to get T_GenCamDcc
   T_GenCamDcc= uniform(T_GenCamMin, T_GenCamMax,0);

   std::cout<<"  T_GenCamDcc: "<< T_GenCamDcc <<endl;

   if (T_GenCamDcc < T_GenCamMin || T_GenCamDcc > T_GenCamMax) {
           throw cRuntimeError("GenCamDel is out of bounds");
   }



  basicCamContainer * lastmcam = carCamGeneration(host, mobility);
   sendCam(getSelfAddress());

   std::cout<<"  checkPosition(curPosition,curPosition):       "<< checkPosition(curPosition,prePosition) <<endl;
   std::cout<<"  checkSpeed(curSpeed,preSpeed):                "<< checkSpeed(curSpeed,preSpeed) <<endl;
   std::cout<<"  checkHeading(curPosition,curPosition):        "<< checkHeading(curPosition,prePosition) <<endl;

   if(T_GenCamDel>T_GenCamMax) {

         basicCamContainer * lastmcam = carCamGeneration(host, mobility);
         sendCam(getSelfAddress());

     }
   else if(T_GenCamDel>=T_GenCamDcc)
   {
       if(checkPosition(curPosition,prePosition) && checkSpeed(curSpeed,preSpeed)&& checkHeading(curPosition,prePosition) ){
            basicCamContainer * lastmcam = carCamGeneration(host, mobility);
            sendCam(getSelfAddress());
      }
        else {
             positionTableforMobilityCam.setPosition(selfAddress,curPosition,curSpeed);
      }
   }

    std::cout<<"  CheckCamCondition               over over over               " <<endl;
}

// checke the conditions every 100ms
void UDPBasicAppCam::CheckCamConditionTimer()
{
    IPvXAddress selfAddress = getSelfAddress();
    CheckCamCondition(selfAddress);
    std::cout<<"  enter into  CheckCamConditionTimer() " <<endl;
    scheduleCamTimer();
}


void UDPBasicAppCam::scheduleCamTimer()
{
    std::cout<<"   scheduleCamTimer " <<endl;

    //camInterval=100ms

    scheduleAt(simTime() + camInterval, camTimer);
}


void UDPBasicAppCam::handleMessage(cMessage * msg)
{
    std::cout<<"  before handleMessage:" <<endl;
    UDPBasicApp::handleMessageWhenUp(msg);

    if(msg=camTimer){
        std::cout<<" handlemessage: CheckCamConditionTimer " <<endl;

        CheckCamConditionTimer();
    } else if (msg->getKind() == UDP_I_DATA) {
            // process incoming packet
            processPacket(PK(msg));
    }

}


void UDPBasicAppCam::sendCam( IPvXAddress selfAddress ){
    std::cout<<"  sendCam            begin begin begin  " <<endl;

    selfAddress = getSelfAddress();
    basicCamContainer * sendcam;

    sendcam->setAddress(selfAddress);
    sendcam->setPosition(positionTableforMobilityCam.getPosition(selfAddress));
    sendcam->setSpeed(positionTableforMobilityCam.getSpeed(selfAddress));
    sendcam->setCreateTime(positionTableforMobilityCam.getCreateTime(selfAddress));

    Enter_Method_Silent();
    char msgName[32];
    sprintf(msgName, "UDPBasicAppCam-%d",numCamSent);

    std::cout <<"msgName:  "<<msgName<<endl;

    cPacket *sendcampayload = new cPacket(sendcam->getName());
    sendcampayload->setByteLength(par("messageLength").longValue());

    PDUHeader *createPduHeader=CreatePduHeader();
    sendcampayload->encapsulate(createPduHeader);

    destAddr = IPv4Address::ALLONES_ADDRESS;

//   IPvXAddress destAddr=chooseDestAddr();


    emit(sentCamSignal, sendcampayload);



    sendBroadcast(destAddr,sendcampayload);

    numCamSent++;

}

PDUHeader * UDPBasicAppCam::CreatePduHeader()
{
    PDUHeader *pduHeader =new PDUHeader();

   pduHeader->setBitLength(10);
   pduHeader->setProtocolVersion(1);
   pduHeader->setMessageID(numCamSent);
   pduHeader->setStationID(host->getIndex());
   pduHeader->addPar("sourceId") = getId();

   return pduHeader;
}


void UDPBasicAppCam::initialConfiguration()
{
   socket.setOutputGate(gate("udpOut"));
// socket.bind(localPort);

    const char *destAddrs = par("destAddresses");
    cStringTokenizer tokenizer(destAddrs);
    const char *token;
    bool excludeLocalDestAddresses = par("excludeLocalDestAddresses").boolValue();
    if (par("setBroadcast").boolValue())
        socket.setBroadcast(true);

    if (strcmp(par("outputInterface").stringValue(),"") != 0)
    {
        IInterfaceTable* ift = InterfaceTableAccess().get();
        InterfaceEntry *ie = ift->getInterfaceByName(par("outputInterface").stringValue());
        if (ie == NULL)
            throw cRuntimeError(this, "Invalid output interface name : %s",par("outputInterface").stringValue());
        outputInterface = ie->getInterfaceId();
    }

    outputInterfaceMulticastBroadcast.clear();
    if (strcmp(par("outputInterfaceMulticastBroadcast").stringValue(),"") != 0)
    {
        IInterfaceTable* ift = InterfaceTableAccess().get();
        const char *ports = par("outputInterfaceMulticastBroadcast");
        cStringTokenizer tokenizer(ports);
        const char *token;
        while ((token = tokenizer.nextToken()) != NULL)
        {
            if (strstr(token, "ALL") != NULL)
            {
                for (int i = 0; i < ift->getNumInterfaces(); i++)
                {
                    InterfaceEntry *ie = ift->getInterface(i);
                    if (ie->isLoopback())
                        continue;
                    if (ie == NULL)
                        throw cRuntimeError(this, "Invalid output interface name : %s", token);
                    outputInterfaceMulticastBroadcast.push_back(ie->getInterfaceId());
                }
            }
            else
            {
                InterfaceEntry *ie = ift->getInterfaceByName(token);
                if (ie == NULL)
                    throw cRuntimeError(this, "Invalid output interface name : %s", token);
                outputInterfaceMulticastBroadcast.push_back(ie->getInterfaceId());
            }
        }
    }
#ifdef WITH_IPv4
    IRoutingTable *rt = RoutingTableAccess().getIfExists();
#endif
#ifdef WITH_IPv6
    RoutingTable6 *rt6 = RoutingTable6Access().getIfExists();
#endif

    while ((token = tokenizer.nextToken()) != NULL)
    {
        if (strstr(token, "Broadcast") != NULL)
            destAddresses.push_back(IPv4Address::ALLONES_ADDRESS);
        else
        {
            IPvXAddress addr = IPvXAddressResolver().resolve(token);
#ifdef WITH_IPv4
            if (excludeLocalDestAddresses && rt && rt->isLocalAddress(addr.get4()))
                continue;
#endif
#ifdef WITH_IPv6
            if (excludeLocalDestAddresses && rt6 && rt6->isLocalAddress(addr.get6()))
                continue;
#endif
            destAddresses.push_back(addr);
        }
    }
}


bool UDPBasicAppCam::sendBroadcast(const IPvXAddress &dest, cPacket *pkt)
{

   dest.get4()=IPv4Address::ALLONES_ADDRESS;

    if (!outputInterfaceMulticastBroadcast.empty() && (dest.isMulticast() || (!dest.isIPv6() && dest.get4() == IPv4Address::ALLONES_ADDRESS)))
    {
        for (unsigned int i = 0; i < outputInterfaceMulticastBroadcast.size(); i++)
        {
            UDPSocket::SendOptions options;
            options.outInterfaceId = outputInterfaceMulticastBroadcast[i]+1;
            std::cout<<"  options.outInterfaceId "<<options.outInterfaceId <<endl;
            if (outputInterfaceMulticastBroadcast.size() - i > 1)
            {
                socket.sendTo(pkt->dup(), dest, destPort, &options);
                            std::cout<<"  sendBroadcast test 1111111111111111111111111 " <<endl;
            }
            else
            {
                socket.sendTo(pkt, dest, destPort, &options);
                std::cout<<"  sendBroadcast test 2222222222222222222222222222222222222222 " <<endl;
            }

        }
        return true;
    }
    return false;
}




